-- ------------------------------------------------------------------------------
-- rp:Friends
-- by Oraibi, Moon Guard (US) server
--
-- Ora's twitter: http://twitter.com/oraibimoonguard
--
-- ------------------------------------------------------------------------------
--
-- This work is licensed under the Creative Commons Attribution 4.0 International
-- (CC BY 4.0) license. To view a copy of the license, visit
--
--     https://creativecommons.org/licenses/by/4.0/
--
-- This license is acceptable for Free Cultural Works.
--
-- You are free to:
--
-- Share -- copy and redistribute the material in any medium or format
-- Adapt -- remix, transform, and build upon the material
--         for any purpose, even commercially.
-- 
-- The licensor cannot revoke these freedoms as long as you follow the 
-- license terms.
-- 
-- Under the following terms:
-- Attribution -- You must give appropriate credit, provide a link to the
--         license, and indicate if changes were made. You may do
--         so in any reasonable manner, but not in any way that
--         suggests the licensor endorses you or your use.
--
-- No additional restrictions -- You may not apply legal terms or
--         technological measures that legally restrict others from doing
--         anything the license permits.
--
-- Notices:
--
-- You do not have to comply with the license for elements of the material
-- in the public domain or where your use is permitted by an applicable
-- exception or limitation.
--
-- No warranties are given. The license may not give you all of the
-- permissions necessary for your intended use. For example, other rights
-- such as publicity, privacy, or moral rights may limit how you use the
-- material.
-- ------------------------------------------------------------------------------

  -- Utils-Config: Utilities for reading and registering configuration values

local RP_FRIENDS     = RP_FRIENDS;

table.insert(RP_FRIENDS.startup, function(state)
  local Const      = RP_FRIENDS.const;
  local Utils      = RP_FRIENDS.utils;
  local RP_FriendsDB = _G.RP_FriendsDB;
  
  if not Utils.config then Utils.config = {} end;
  local Config = Utils.config;
  
  local Default = Const.CONFIG.DEFAULTS;
  
  -- fixes for when we've had changes in config options
  local Fix = {
    ALPHA           = function(value) if value > 1     then value = value/100                                           end return value end,
 -- CHANGES_MESSAGE = function(value) if value == "ALWAYS" then value = true elseif value == "NEVER" then value = false end return value end,
  };

  for key, _ in pairs(Default)
  do  local db = RP_FriendsDB.settings;
      if not db[key] then db[key] = {} end;
      if   db[key].value ~= nil then db[key].value = db[key].value 
      else db[key].value = Default[key];
      end;
      if Fix[key] then db[key].value = Fix[key](db[key].value); end;
      db[key].default = Default[key];
      db[key].init = nil;
  end;
         
        -- return the value of a config setting
  local function getConfig(key) if not key then return nil; end;
    if  RP_FriendsDB.settings[key] then return RP_FriendsDB.settings[key].value else return nil end;
  end;  
  
  local function setConfig(key, value) if not key then return nil; end;
    if   RP_FriendsDB.settings[key] then RP_FriendsDB.settings[key].value = value; return value; else return nil; end; -- if;
  end; -- function
  
  local function resetConfig(key) if not key then return nil; end; 
    if    RP_FriendsDB.settings[key]
    then RP_FriendsDB.settings[key].value = RP_FriendsDB.settings[key].default;
         return RP_FriendsDB.settings[key].default;
    else return nil;
    end; --if
  end; -- function
  
  local function getDefault(key) if not key then return nil; end;
    if RP_FriendsDB.settings[key]
    then return RP_FriendsDB.settings[key].default or Default[key];
    else return nil;
    end;
  end; -- function
  
  -- Functions available under RP_FRIENDS.utils.config
  --
  Config.get        = getConfig;
  Config.set        = setConfig;
  Config.reset      = resetConfig;
  Config.default    = getDefault;

  -- keybindings
  BINDING_HEADER_RP_FRIENDS              = RP_FRIENDS.utils.locale.loc("APP_NAME");

  return "config"

end); -- startup function
