local RP_FRIENDS = RP_FRIENDS;

table.insert(RP_FRIENDS.startup, function(state)

  local loc          = RP_FRIENDS.utils.locale.loc;
  local Config       = RP_FRIENDS.utils.config;
  local Grid         = RP_FRIENDS.grid;

  local CategoryPicker = CreateFrame('Frame', 'RP_FRIENDS_CategoryPicker', RP_FRIENDS.frame)
  RP_FRIENDS.frame.category = CategoryPicker;

  CategoryPicker:SetSize(307, 109);
  CategoryPicker:SetBackdrop(RP_FRIENDS.const.BACKDROP.BLIZZTOOLTIPTHIN);
  CategoryPicker:SetBackdropColor(0, 0, 0, 0.75);
  CategoryPicker.buttons = {};
  CategoryPicker:SetFrameStrata('DIALOG');

  CategoryPicker.title = CategoryPicker:CreateFontString(nil, 'OVERLAY', 'GameFontNormalLarge');
  CategoryPicker.title:SetText(' ');
  CategoryPicker.title:SetPoint("TOPLEFT", CategoryPicker, "TOPLEFT", 8, -8);
  CategoryPicker.title:SetPoint("TOPRIGHT", CategoryPicker, "TOPRIGHT", -58, -8);
  CategoryPicker.title:SetWordWrap(false);
  CategoryPicker.title:SetJustifyH("LEFT");
  CategoryPicker.title:SetJustifyV("TOP");
  
  CategoryPicker.closeButton = CreateFrame('Frame', nil, CategoryPicker);
  CategoryPicker.closeButton.texture = CategoryPicker.closeButton:CreateTexture();
  CategoryPicker.closeButton.texture:SetTexture("Interface\\RAIDFRAME\\ReadyCheck-Ready");
  CategoryPicker.closeButton.texture:SetAllPoints(CategoryPicker.closeButton);
  CategoryPicker.closeButton:SetScript("OnMouseDown", function(self) RP_FRIENDS.frame.category:Hide() end);
  CategoryPicker.closeButton:SetScript("OnEnter",
    function(self)
      GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
      GameTooltip:ClearLines();
      GameTooltip:AddLine("|cff11dd11Finished|r");
      GameTooltip:AddLine("Click the checkmark when you are done setting categories.", 0.7, 0.7, 0.7, true)
      GameTooltip:Show();
    end);
  CategoryPicker.closeButton:SetScript("OnLeave", function(self) GameTooltip:Hide() end);
  CategoryPicker.closeButton:SetSize(24, 24);
  CategoryPicker.closeButton:SetPoint("TOPRIGHT", CategoryPicker, "TOPRIGHT", -4, -4);

  anchor, lastElement, relAnchor, xOffset, yOffset = "TOPLEFT", CategoryPicker.title, "BOTTOMLEFT", 0, -8;
  for cat = 0, 15
  do  local CategoryButton = CreateFrame('Frame', nil, CategoryPicker);
      CategoryButton.i = cat;
      CategoryButton:SetSize(32, 32);
      CategoryButton.text = CategoryButton:CreateFontString(nil, "OVERLAY", "GameFontNormalHuge3");
      CategoryButton.text:SetAllPoints(CategoryButton);
      CategoryButton.text:SetText(RP_FRIENDS.const.icons.SMALL[Config.get(string.format("CAT%X_ICON", i))].icon)
      CategoryButton:SetPoint(anchor, lastElement, relAnchor, xOffset, yOffset);
      if     cat % 8 == 7
      then   anchor, lastElement, relAnchor, xOffset, yOffset = "TOPLEFT", lastLine, "BOTTOMLEFT", 0, -5;
      elseif cat % 8 == 0
      then   anchor, lastElement, relAnchor, xOffset, yOffset = "LEFT", CategoryButton, "RIGHT", 5, 0;
             lastLine = CategoryButton;
      else  anchor, lastElement, relAnchor, xOffset, yOffset = "LEFT", CategoryButton, "RIGHT", 5, 0;
      end;
      CategoryPicker.buttons[cat] = CategoryButton;
      CategoryButton.bg = CategoryPicker:CreateTexture();
      CategoryButton.bg:SetColorTexture(1, 1, 1, 0.25);
      CategoryButton.bg:SetAllPoints(CategoryButton);
      CategoryButton:SetScript("OnMouseDown",
        function(self)
          if   self.value 
          then self:GetParent().OnUnselect(self:GetParent(), self.i, self:GetParent().f);
               self.bg:Hide();
               self.value = false;
          else self:GetParent().OnSelect(self:GetParent(), self.i, self:GetParent().f);
               self.bg:Show();
               self.value = true;
          end;
          RP_FRIENDS.frame.display:Update("datashowicons", "category picker icons chosen")
        end);
      CategoryButton:SetScript("OnEnter",
        function(self)
          GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
          GameTooltip:ClearLines();
          GameTooltip:AddLine(Config.get(string.format("CAT%X", self.i)));
          GameTooltip:AddLine(Config.get(string.format("CAT%X_DESC"), self.i), 1, 1, 1, true);
          GameTooltip:AddLine(' ');
          GameTooltip:Show()
        end);
      CategoryButton:SetScript("OnLeave", function(self) GameTooltip:Hide() end);
  end;

  CategoryPicker:SetMovable(true);
  CategoryPicker:SetPoint("CENTER", RP_FRIENDS.frame.display, "CENTER", 0, 0);
  CategoryPicker:EnableMouse(true);
  CategoryPicker:RegisterForDrag("LeftButton");
  CategoryPicker:SetScript("OnDragStart", function(self) self:StartMoving() SetCursor("UI_MOVE_CURSOR") GameTooltip:Hide() end);
  CategoryPicker:SetScript("OnDragStop", function(self) self:StopMovingOrSizing() SetCursor(nil) end);
  CategoryPicker:SetClampedToScreen(true);
  tinsert(UISpecialFrames, 'RP_FRIENDS_CategoryPicker');

  function CategoryPicker:SetTitle(title) self.title:SetText(title) end;

  function CategoryPicker:HiliteSelected(cats)
    for i = 0, 15, 1
    do  self.buttons[i].text:SetText(RP_FRIENDS.const.icons.SMALL[Config.get(string.format("CAT%X_ICON", i))].icon)
        self.buttons[i].bg:SetShown(bit.band(cats,2 ^ i) > 0);
        self.buttons[i].value = (bit.band(cats,2 ^ i) > 0);
    end;
  end;

  CategoryPicker:SetScript("OnHide", function(self) self.f = nil; end);

  function CategoryPicker:SetFriendCategory(f)
    self.title:SetText(
         f.bnet_battleTag     and (BATTLENET_FONT_COLOR_CODE             .. f.bnet_battleTag     .. "|r")
      or f.rp_name_nocolor    and ("|cff" .. Config.get("COLOR_RP")      .. f.rp_name_nocolor    .. "|r")
      or f.acct_characterName and ("|cff" .. Config.get("COLOR_DEFAULT") .. f.acct_characterName .. "|r")
      or "Friend");
    self.f = f;
  
    self:HiliteSelected(RP_FRIENDS.database:GetCategories(f))
    self.OnSelect   = 
      function(self, i, f) 
        RP_FRIENDS.database:AddCategory(f, i);    
      end;
    self.OnUnselect = 
      function(self, i, f) 
        RP_FRIENDS.database:RemoveCategory(f, i); 
      end;
    self:Show();
  end;

  return "category";
end);  
