-- ------------------------------------------------------------------------------
-- rp:Friends
-- by Oraibi, Moon Guard (US) server
--
-- Ora's twitter: http://twitter.com/oraibimoonguard
--
-- ------------------------------------------------------------------------------
--
-- This work is licensed under the Creative Commons Attribution 4.0 International
-- (CC BY 4.0) license. To view a copy of the license, visit
--
--     https://creativecommons.org/licenses/by/4.0/
--
-- This license is acceptable for Free Cultural Works.
--
-- You are free to:
--
-- Share -- copy and redistribute the material in any medium or format
-- Adapt -- remix, transform, and build upon the material
--         for any purpose, even commercially.
-- 
-- The licensor cannot revoke these freedoms as long as you follow the 
-- license terms.
-- 
-- Under the following terms:
-- Attribution -- You must give appropriate credit, provide a link to the
--         license, and indicate if changes were made. You may do
--         so in any reasonable manner, but not in any way that
--         suggests the licensor endorses you or your use.
--
-- No additional restrictions -- You may not apply legal terms or
--         technological measures that legally restrict others from doing
--         anything the license permits.
--
-- Notices:
--
-- You do not have to comply with the license for elements of the material
-- in the public domain or where your use is permitted by an applicable
-- exception or limitation.
--
-- No warranties are given. The license may not give you all of the
-- permissions necessary for your intended use. For example, other rights
-- such as publicity, privacy, or moral rights may limit how you use the
-- material.
-- ------------------------------------------------------------------------------

local RP_FRIENDS = RP_FRIENDS;

table.insert(RP_FRIENDS.startup, function(state)

  -- libraries
  local Utils = RP_FRIENDS.utils;
  local loc = RP_FRIENDS.utils.locale.loc;

  RP_FRIENDS.frame:RegisterEvent("PLAYER_ENTERING_WORLD");                   -- when player logs on and everything is ready to go

  -- unused events: these are set on the default blizzard friend list
  --
  -- RP_FRIENDS.frame:RegisterEvent("BN_BLOCK_LIST_UPDATED");        -- we don't handle BN blocking
  -- RP_FRIENDS.frame:RegisterEvent("GROUP_JOINED");                 -- these are for the invite-to-group functions which we don't handle
  -- RP_FRIENDS.frame:RegisterEvent("GROUP_LEFT");                   -- these are for the invite-to-group functions which we don't handle
  -- RP_FRIENDS.frame:RegisterEvent("GROUP_ROSTER_UPDATE");          -- these are for the invite-to-group functions which we don't handle
  -- RP_FRIENDS.frame:RegisterEvent("GROUP_ROSTER_UPDATE");          -- these are for the invite-to-group functions which we don't handle
  -- RP_FRIENDS.frame:RegisterEvent("GUILD_ROSTER_UPDATE");          -- these are for the invite-to-group functions which we don't handle
  -- RP_FRIENDS.frame:RegisterEvent("PARTY_REFER_A_FRIEND_UPDATED"); -- we don't handle RAF
  -- RP_FRIENDS.frame:RegisterEvent("IGNORELIST_UPDATE");            -- we don't handle ignores
  -- RP_FRIENDS.frame:RegisterEvent("SOCIAL_QUEUE_UPDATE");          -- we don't handle social queueing
  -- RP_FRIENDS.frame:RegisterEvent("SPELL_UPDATE_COOLDOWN");        -- we don't handle summoning people
  -- RP_FRIENDS.frame:RegisterEvent("WHO_LIST_UPDATE");              -- we don't handle who lists

  RP_FRIENDS.frame:SetScript("OnEvent",
    function (self, event, arg, ...)
      if     event == "PLAYER_ENTERING_WORLD" 
      then   local isInitialLogin, isReload = ...
             if isInitialLogin or isReload
             then -- things that should only run once

                  -- register for events 
                  self:RegisterEvent("BN_CONNECTED");                      -- when @player connects to BN
                  self:RegisterEvent("BN_DISCONNECTED");                   -- when @player disconnects from BN
                  self:RegisterEvent("BN_CUSTOM_MESSAGE_CHANGED");         -- when a friend changes their BN message
                                                                           --   Payload: friendIndex
                  self:RegisterEvent("BN_CUSTOM_MESSAGE_LOADED");          -- when a BN message is loaded
                  self:RegisterEvent("BN_FRIEND_INFO_CHANGED");            -- when a BN friend changes their info
                                                                           --   Payload: friendIndex
                  self:RegisterEvent("BN_FRIEND_LIST_SIZE_CHANGED");       -- when @player adds or removes a BN friend, or someone logs on or disconnects
                                                                           --   Payload: accountID
                  self:RegisterEvent("BN_INFO_CHANGED");                   -- when my own battlenet info changes
                  self:RegisterEvent("PLAYER_FLAGS_CHANGED");              -- "This event fires when a Unit's flags change (e.g. /afk, /dnd)."
                                                                           -- Payload: "unitTarget (string)"
                  self:RegisterEvent("BATTLETAG_INVITE_SHOW");             -- 
                  self:RegisterEvent("BN_FRIEND_ACCOUNT_OFFLINE");         --
                  self:RegisterEvent("BN_FRIEND_ACCOUNT_ONLINE");          --
                  self:RegisterEvent("BN_FRIEND_INVITE_ADDED");            -- 
                  self:RegisterEvent("BN_FRIEND_INVITE_LIST_INITIALIZED"); -- 
                  self:RegisterEvent("BN_FRIEND_INVITE_REMOVED");          -- 
                  self:RegisterEvent("FRIENDLIST_UPDATE");                       

                  --UI creation
                  RP_FRIENDS.frame.button:CreateButton();                        -- create the minimap button

                  -- apply config settings
                  self:ShowButtonBar();                                    -- makes the extra buttons (Overwatch, etc) show/hide
                  RP_FRIENDS.frame:ShrinkStatusBar();                      -- hides the friend count if RP_FRIENDS.frame is too small
                  RP_FRIENDS.frame:UpdateMyOwnStatus();                    -- set the toggle of the IC/OOC button if state.mrp or state.trp3
                  
                  -- friendDB
                  RP_FRIENDS.database:UpdateAll(nil, true, event); 

             else -- things that should run every time PLAYER_ENTERING_WORLD fires
                  RP_FRIENDS.database:UpdateAll(nil, false, event);
                  self:ShowButtonBar();
             end;

      elseif event == "BN_CONNECTED" or event == "BN_DISCONNECTED"
      then   RP_FRIENDS.database:UpdateAll(nil, false, event, arg);
             self:ShowButtonBar();
      elseif (event == "BATTLETAG_INVITE_SHOW" or event == "BN_FRIEND_INVITE_ADDED" or event == "BN_FRIEND_INVITE_LIST_INITIALIZE"
          or event == "BN_FRIEND_INVITE_REMOVED") and self:IsShown()
        then self:InviteGlow();

      elseif event == "BN_FRIEND_LIST_SIZE_CHANGED" or event == "BN_CUSTOM_MESSAGE_CHANGED"
      then   RP_FRIENDS.database:UpdateAll("bn", true, event, arg)

      elseif event == "FRIENDLIST_UPDATE"
      then   RP_FRIENDS.database:UpdateAll("server", true, event, arg)

      elseif event == "BN_FRIEND_ACCOUNT_OFFLINE" or event == "BN_FRIEND_ACCOUNT_ONLINE" or 
             event == "BN_FRIEND_INFO_CHANGED"    
      then   
             if arg then RP_FRIENDS.database:GetBNFriend(arg);
             else RP_FRIENDS.database:UpdateAll("bn", true, event, arg)
             end;

      end; --if

    end); -- setscript ##

  table.insert(msp.callback.received, 
    function(u) 
      local friendID = RP_FRIENDS.database:IsTracked(u);
      if   friendID
      then RP_FRIENDS.database:UpdatePlayerMSP(friendID)
           RP_FRIENDS.frame.display:Update("datashow", "msp.callback.received for friend " .. friendID) 
      end; 
      if u == UnitName('player') .. "-" .. GetRealmName():gsub(" ", "") then RP_FRIENDS.frame:UpdateMyOwnStatus() end;
    end); -- callback

  if state.trp3
  then 
    TRP3_API.module.registerModule( { -- let trp3 know we exist and set up our callback for new profile received
      ["name"]        = loc("APP_NAME"),
      ["description"] = loc("APP_DESC"),
      ["version"]     = loc("APP_VERSION"),
      ["id"]          = loc("APP_ID"),
      ["minVersion"]  = 3,
      ["autoEnable"]  = true,
       }); -- closes the addon registering table and the function params
   end;
  --------------------------

  return "events";
end);
