local RP_FRIENDS = RP_FRIENDS;

table.insert(RP_FRIENDS.startup, function(state)

  local loc          = RP_FRIENDS.utils.locale.loc;
  local Config       = RP_FRIENDS.utils.config;
  local Grid         = RP_FRIENDS.grid;

  local CategoryPicker = CreateFrame('Frame', 'RP_FRIENDS_CategoryPicker', RP_FRIENDS.frame)
  RP_FRIENDS.frame.category = CategoryPicker;

  CategoryPicker:SetSize(307, 109);
  CategoryPicker:SetBackdrop(RP_FRIENDS.const.BACKDROP.BLIZZTOOLTIPTHIN);
  CategoryPicker:SetBackdropColor(0, 0, 0, 0.75);
  CategoryPicker.categories = {};
  CategoryPicker:SetFrameStrata('DIALOG');

  CategoryPicker.title = CategoryPicker:CreateFontString(nil, 'OVERLAY', 'GameFontNormalLarge');
  CategoryPicker.title:SetText(' ');
  CategoryPicker.title:SetPoint("TOPLEFT", CategoryPicker, "TOPLEFT", 8, -8);
  CategoryPicker.title:SetPoint("TOPRIGHT", CategoryPicker, "TOPRIGHT", -58, -8);
  CategoryPicker.title:SetWordWrap(false);
  CategoryPicker.title:SetJustifyH("LEFT");
  CategoryPicker.title:SetJustifyV("TOP");
  
  CategoryPicker.closeButton = CreateFrame('Frame', nil, CategoryPicker);
  CategoryPicker.closeButton.texture = CategoryPicker.closeButton:CreateTexture();
  CategoryPicker.closeButton.texture:SetTexture("Interface\\RAIDFRAME\\ReadyCheck-Ready");
  CategoryPicker.closeButton.texture:SetAllPoints(CategoryPicker.closeButton);
  CategoryPicker.closeButton:SetScript("OnMouseDown", function(self) RP_FRIENDS.frame.category:Hide() end);
  CategoryPicker.closeButton:SetScript("OnEnter",
    function(self)
      GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
      GameTooltip:ClearLines();
      GameTooltip:AddLine("|cff11dd11Finished|r");
      GameTooltip:AddLine("Click the checkmark when you are done setting categories.", 0.7, 0.7, 0.7, true)
      GameTooltip:Show();
    end);
  CategoryPicker.closeButton:SetScript("OnLeave", function(self) GameTooltip:Hide() end);
  CategoryPicker.closeButton:SetSize(24, 24);
  CategoryPicker.closeButton:SetPoint("TOPRIGHT", CategoryPicker, "TOPRIGHT", -4, -4);

  anchor, lastElement, relAnchor, xOffset, yOffset = "TOPLEFT", CategoryPicker.title, "BOTTOMLEFT", 0, -8;
  for cat = 0, 15
  do  local hex = string.format("%x", cat):upper();
      local CategoryFrame = CreateFrame('Frame', nil, CategoryPicker);
      CategoryFrame:SetSize(32, 32);
      CategoryFrame.text = CategoryFrame:CreateFontString(nil, "OVERLAY", "GameFontNormalHuge3");
      CategoryFrame.text:SetAllPoints(CategoryFrame);
      CategoryFrame.text:SetText(hex);
      CategoryFrame:SetPoint(anchor, lastElement, relAnchor, xOffset, yOffset);
      if     cat % 8 == 7
      then   anchor, lastElement, relAnchor, xOffset, yOffset = "TOPLEFT", lastLine, "BOTTOMLEFT", 0, -5;
      elseif cat % 8 == 0
      then   anchor, lastElement, relAnchor, xOffset, yOffset = "LEFT", CategoryFrame, "RIGHT", 5, 0;
             lastLine = CategoryFrame;
      else  anchor, lastElement, relAnchor, xOffset, yOffset = "LEFT", CategoryFrame, "RIGHT", 5, 0;
      end;
      CategoryFrame.hex = hex;
      CategoryPicker.categories[hex] = CategoryFrame;
      CategoryFrame.bg = CategoryPicker:CreateTexture();
      CategoryFrame.bg:SetColorTexture(1, 1, 1, 0.25);
      CategoryFrame.bg:SetAllPoints(CategoryFrame);
      CategoryFrame:SetScript("OnMouseDown",
        function(self)
          if   self.value 
          then self:GetParent().OnUnselect(self:GetParent(), self.hex, self:GetParent().f);
               self.bg:Hide();
               self.value = false;
          else self:GetParent().OnSelect(self:GetParent(), self.hex, self:GetParent().f);
               self.bg:Show();
               self.value = true;
          end;
          RP_FRIENDS.frame.display:Update("datashowicons", "category picker icons chosen")
        end);
      CategoryFrame:SetScript("OnEnter",
        function(self)
          GameTooltip:SetOwner(self, "AMCHOR_RIGHT");
          GameTooltip:ClearLines();
          GameTooltip:AddLine(Config.get("CAT" .. self.hex));
          GameTooltip:AddLine(Config.get("CAT" .. self.hex .. "_DESC"), 1, 1, 1, true);
          GameTooltip:AddLine(' ');
          GameTooltip:Show()
        end);
      CategoryFrame:SetScript("OnLeave", function(self) GameTooltip:Hide() end);
  end;

  CategoryPicker:SetMovable(true);
  CategoryPicker:SetPoint("CENTER", RP_FRIENDS.frame.display, "CENTER", 0, 0);
  CategoryPicker:EnableMouse(true);
  CategoryPicker:RegisterForDrag("LeftButton");
  CategoryPicker:SetScript("OnDragStart", function(self) self:StartMoving() SetCursor("UI_MOVE_CURSOR") GameTooltip:Hide() end);
  CategoryPicker:SetScript("OnDragStop", function(self) self:StopMovingOrSizing() SetCursor(nil) end);
  CategoryPicker:SetClampedToScreen(true);
  tinsert(UISpecialFrames, 'RP_FRIENDS_CategoryPicker');

  function CategoryPicker:SetTitle(title) self.title:SetText(title) end;

  function CategoryPicker:HiliteSelected(cats)
    local cats = Config.get("CATEGORY_FILTER") 
    for i, CategoryFrame in pairs(self.categories)
    do  CategoryFrame.text:SetText(RP_FRIENDS.const.icons.SMALL[Config.get("CAT" .. i .. "_ICON")].icon)
        CategoryFrame.bg:SetShown(bit.band(cats,2 ^ i));
        CategoryFrame.value = bit.band(cats,2 ^ i);
    end;
  end;

  CategoryPicker:SetScript("OnHide", function(self) self.f = nil; end);

  function CategoryPicker:SetFriendCategory(f)
    self.title:SetText(
         f.bnet_battleTag     and (BATTLENET_FONT_COLOR_CODE             .. f.bnet_battleTag     .. "|r")
      or f.rp_name_nocolor    and ("|cff" .. Config.get("COLOR_RP")      .. f.rp_name_nocolor    .. "|r")
      or f.acct_characterName and ("|cff" .. Config.get("COLOR_DEFAULT") .. f.acct_characterName .. "|r")
      or "Friend");
    self.f = f;
  
    self:HiliteSelected(RP_FRIENDS.database:GetCategories(f))
    self.OnSelect   = function(self, hex, f) RP_FRIENDS.database:AddCategory(f, hex);    end;
    self.OnUnselect = function(self, hex, f) RP_FRIENDS.database:RemoveCategory(f, hex); end;
    self:Show();
  end;

  return "category";
end);  
