-- ------------------------------------------------------------------------------
-- rp:Friends
-- by Oraibi, Moon Guard (US) server
--
-- Ora's twitter: http://twitter.com/oraibimoonguard
--
-- ------------------------------------------------------------------------------
--
-- This work is licensed under the Creative Commons Attribution 4.0 International
-- (CC BY 4.0) license. To view a copy of the license, visit
--
--     https://creativecommons.org/licenses/by/4.0/
--
-- This license is acceptable for Free Cultural Works.
--
-- You are free to:
--
-- Share -- copy and redistribute the material in any medium or format
-- Adapt -- remix, transform, and build upon the material
--         for any purpose, even commercially.
-- 
-- The licensor cannot revoke these freedoms as long as you follow the 
-- license terms.
-- 
-- Under the following terms:
-- Attribution -- You must give appropriate credit, provide a link to the
--         license, and indicate if changes were made. You may do
--         so in any reasonable manner, but not in any way that
--         suggests the licensor endorses you or your use.
--
-- No additional restrictions -- You may not apply legal terms or
--         technological measures that legally restrict others from doing
--         anything the license permits.
--
-- Notices:
--
-- You do not have to comply with the license for elements of the material
-- in the public domain or where your use is permitted by an applicable
-- exception or limitation.
--
-- No warranties are given. The license may not give you all of the
-- permissions necessary for your intended use. For example, other rights
-- such as publicity, privacy, or moral rights may limit how you use the
-- material.
-- ------------------------------------------------------------------------------

local RP_FRIENDS = RP_FRIENDS;

table.insert(RP_FRIENDS.startup, function(state)

  -- libraries
  --
  local LibRealmInfo  = LibStub("LibRealmInfo");

  -- data structure 
  --
  RP_FRIENDS.database = {
    data = {},                   -- contains all f-records
    hash   = {                   -- indexes to f-records
      bnet_battleTag     = {},   -- database.hash.bnet_battleTag["Oraibi#1617"] = { "Oraibi#1617-WoW", "Oraibi-MoonGuard", "Oraibi#1617-App", "Oraibi#1617-MSAp" }
      acct_unitID        = {},   -- database.hash.acct_unitID["Oraibi-MoonGuard"] = { btag = "Oraibi#1617-WoW", unitID = "Oraibi-MoonGuard" }
      },
    btags = {},
    msp_fields = {
      "CO",  -- rp_oocinfo
      "CU",  -- rp_curr
      "FC",  -- rp_status
      "IC",  -- rp_icon
      "NA",  -- rp_name
      "NI",  -- rp_nick
      "NT",  -- rp_title
      "PX",  -- rp_honorific
      "RA",  -- rp_race
      "RC",  -- rp_class
    },
  };

  local database = RP_FRIENDS.database;
  local BTAG_FIELDS  = RP_FRIENDS.grid.FIELDS.BTAG; 
  local ACCT_FIELDS  = RP_FRIENDS.grid.FIELDS.ACCT; 
  local REALM_FIELDS = RP_FRIENDS.grid.FIELDS.REALM;
  
  function database:GenerateFriendID(f) 
    return (f.bnet_battleTag and (f.bnet_battleTag) ..  "-" ..  (f.acct_client or f.bnet_client or "DISCONNECT"))
        or (f.acct_characterName) .. "-" ..(f.acct_realmName or ""):gsub(" ","");
  end; 
    
  function database:IsTracked(unitID)
    local refs = self.hash.acct_unitID[unitID];
    if not refs then return nil end;
    return (refs.btag or refs.unitID) and unitID;
  end;

  function database:ClearFRecord(friendID)
    self.
  end;

  end;
  function database:ClearUnitID(unitID)
    self.hash.acct_unitID = {};
    
  end;

  function database:CleanBNFriend(bnet_bnetIDAccount)
    if not self.btags[bnet_bnetIDAccount] then return end; -- nothing to do
    local clients = {};
    local friendAccounts = BNGetNumFriendGameAccounts(bnet_bnetIDAccount);
    if   friendAccounts > 0
    then for acctNum = 1, friendAccounts
         do local _, _, client = BNGetFriendGameAccountInfo(bnet_bnetIDAccount, acctNum)
            clients[client] = true;
         end;
    end;
    for client, friendID in pairs(self.btags[bnet_bnetIDAccount])
    do  if not clients[client] then self.btags[bnet_bnetIDAccount][client] = nil; self:Clear(friendID) end;
    end;
  end;

  function database:UpdateBNFriends()
    local numBNFriends, numBNFriendsOnline = BNGetNumFriends();
    RP_FRIENDS.frame.statusBar.bn.count:SetText(numBNFriendsOnline .. "/" .. numBNFriends);

    for friendNum = 1, numBNFriends
    do  local bnData = { BNGetFriendInfo(friendNum) };
        local f = {}; 
        for i, field in ipairs(BTAG_FIELDS) do f[field] = bnData[i] end;
        local friendAccounts = BNGetNumFriendGameAccounts(friendNum);
        f.bnet_friendNum = friendNum;
        if    friendAccounts > 0
        then  f.accountList = {};
              for acctNum = 1, friendAccounts
              do  local acctData = { BNGetFriendGameAccountInfo(friendNum, acctNum) };
                  local a = {}; for k, v in pairs(f) do a[k] = v end;
                  for i, field in ipairs(ACCT_FIELDS) do a[field] = acctData[i] end;
                  if   a.acct_client == "WoW" and a.acct_realmName and a.acct_realmName ~= ""
                  then local realmData = { LibRealmInfo:GetRealmInfo(a.acct_realmName) };
                       for i, field in pairs(REALM_FIELDS) do a[field] = realmData[i] end;
                       a.unitID = a.acct_characterName .. "-" .. a.acct_realmName:gsub(" ","");
                  end;
                  self:StoreFriend(a);
                  table.insert(f.accountList, a.acct_client or "Unknown")
              end; -- for a
        else f.bnet_client = "DISCONNECT"; 
             self:StoreFriend(f);
        end; -- if friendAccounts > 0
    end; -- for f
  end; -- function
       
  function database:UpdateServerFriends()
    local numFriends = C_FriendList.GetNumFriends()
    local numFriendsOnline = C_FriendList.GetNumOnlineFriends()
    RP_FRIENDS.frame.statusBar.server.count:SetText(numFriendsOnline .. "/" .. numFriends);

    for friendNum = 1, numFriends
    do  local friendInfo = C_FriendList.GetFriendInfoByIndex(friendNum);
        local f = {
          acct_characterName  = friendInfo.name or "", 
          acct_client         = (friendInfo.connected and "WoW") or "DISCONNECT", 
          acct_class          = friendInfo.className or "",
          acct_zoneName       = friendInfo.area or "",
          acct_charNoteText   = friendInfo.notes or "",
          acct_characterGUID  = friendInfo.guid or "",
          acct_level          = friendInfo.level or 0,
          bnet_isDND          = friendInfo.dnd or false,
          bnet_isAFK          = friendInfo.afk or false,
          bnet_isReferAFriend = friendInfo.referAFriend or false,
          bnet_isOnline       = friendInfo.connected or false,
          acct_friendID       = friendNum,
          acct_faction        = UnitFactionGroup('player'),
          acct_realmName      = RP_FRIENDS.myServer,
        }; 
       f.unitID = f.acct_characterName .. "-" .. f.acct_realmName:gsub(" ","");
       for i, field in pairs(REALM_FIELDS) do f[field] = RP_FRIENDS.cache.myRealm[i] end;
       self:StoreFriend(f);
    end;
  end;

  function database:UpdatePlayerMSP(friendID)
    if not friendID or not self.data[friendID] then return nil end;
    local unitID = self.data[friendID].unitID;

    local profile = msp.char[unitID].field;
    local rpData = 
      { rp_age     = profile.AG, rp_birthplace = profile.BH, rp_class     = profile.RC,  rp_client  = profile.VA,
        rp_curr    = profile.CU, rp_desc       = profile.DE, rp_eyes      = profile.AE,  rp_glance  = profile.PE,
        rp_height  = profile.AH, rp_home       = profile.HH, rp_honorific = profile.PX,  rp_icon    = profile.IC,
        rp_motto   = profile.MO, rp_name       = profile.NA, rp_nick      = profile.NI, 
        rp_oocinfo = profile.CO, rp_title      = profile.NT, rp_race      = profile.RA,  rp_rstatus = profile.RS,
        rp_sliders = profile.PS, rp_status     = profile.FC, rp_tooltip   = profile.TT,  rp_weight  = profile.AW,
        rp_xp      = profile.FR, acct_gender   = profile.GS, 
        rp_name_nocolor = profile.NA and profile.NA:gsub("|cff%x%x%x%x%x%x",""):gsub("|r","") or nil,
        };

    if     state.mrp
    then   local player, realm = unpack(RP_FRIENDS.utils.text.split(unitID:upper(), "-"));
           rpData.rp_noteText = mrpNotes and mrpNotes[realm] and mrpNotes[realm][player]
    elseif state.trp3
    then   rpData.rp_noteText = TRP3_API.profile.getPlayerCurrentProfile().notes and TRP3_Register.character[unitID] and TRP3_Register.character[unitID].profileID
                            and TRP3_API.profile.getPlayerCurrentProfile().notes[TRP3_Register.character[unitID].profileID]
    end;
    for k, v in pairs(rpData) do self:SetFriend(friendID, k, v) end;
  end; -- function

  function database:UpdateAll(request, getMSP, from)
    if request == "server" or not request then self:UpdateServerFriends(); end
    if request == "bn"     or not request then self:UpdateBNFriends();     end;

    RP_FRIENDS.frame.display:Update("colsdatashow", "update all for " .. (request or "all") .. " from " .. (from or "unknown"));
  end; -- function

  -- friend database manipulation ---------------------------------------------------------------------------------------------------------------------------------
  function database:StoreFriend(f)   
    local  friendID = self:GenerateFriendID(f);
    if not self.data[friendID] then self.data[friendID] = {} end;
    for    k, v in pairs(f)             do    self.data[friendID][k] = v                         end;
    for    field, _ in pairs(self.hash) do if f.field then self.hash[field][f.field] = friendID; end; end;
  end;

  function database:AllFriends()                      return self.data; end;
  function database:Get(friendID)                     return self.data[friendID] end;
  function database:Clear(friendID)                   print("Ora, clearing", friendID); self.data[friendID] = nil; end;
  function database:GetValue(friendID, field)         return self.data[friendID] and self.data[friendID][field] or nil; end;
  function database:SetFriend(friendID, field, value) if not self.data[friendID] then self:StoreFriend(friendID, {}); end; self.data[friendID][field] = value;
  end;

  -- --------------------------------------------------------------------------------------------------------------------------------------------------------------

  function database:ShowProfile(f)
    if f.bnet_client == "DISCONNECT" or f.acct_client == "DISCONNECT" then return
    elseif state.mrp  and msp.char[f.unitID] then SlashCmdList.MYROLEPLAY("show " .. f.unitID);
    elseif state.trp3 and msp.char[f.unitID] then SlashCmdList.TOTALRP3("open " .. f.unitID)
    end;
  end;

  -- RP notes updated --------------------------------------------------------------------------------------------------------------------------------------------
  function database:UpdateRPNotes() return end;
    -- for friendID, unitID in pairs(self.tracked)
    -- do local friendData = self:Get(friendID);
    --    if state.mrp
    --    then   local player, realm = unpack(RP_FRIENDS.utils.text.split(unitID:upper(), "-"));
    --           self:SetFriend(friendID, "rp_noteText", mrpNotes and mrpNotes[realm] and mrpNotes[realm][player]);
    --    elseif state.trp3
    --    then   self:SetFriend(friendID, "rp_noteText", 
    --                  TRP3_API.profile.getPlayerCurrentProfile().notes
    --                  and TRP3_Register.character[unitID]
    --                  and TRP3_Register.character[unitID].profileID
    --                  and TRP3_API.profile.getPlayerCurrentProfile().notes[TRP3_Register.character[unitID].profileID]);
    --    end; -- if stat
    -- end; -- for friendID
  -- end


  -- categories code, rewritten from scratch --------------------------------------------------------------------------------------------------------------------
  --
  function database:ExtractCategories(s)
    local  note, xxxx_cats = s:match("^(.+) %[rp(%x%x%x%x)%]"); 
    return note and note or s, note and tonumber(xxxx_cats, 16) or 0 
  end;
   
  function database:PickNote(f) 
    return f.bnet_battleTag and f.bnet_noteText 
        or f.acct_charNoteText 
        or "", 
           f.bnet_battleTag and "BNET" 
        or "ACCT" 
  end;

  function database:GetFriendNoteAndCategories(f)
    local note, kind = self:PickNote(f);
    local stripped_note, bits = self:ExtractCategories(note);
    return f, stripped_note, bits, kind;
  end;

  function database:SetFriendNoteAndCategories(f, note, bits, kind)
    if     bits > 0 and kind == "BNET"
    then   BNSetFriendNote(f.bnet_bnetIDAccount, string.format("%s [rp%x%x%x%x]", note:sub(1, 118), bits))
    elseif bits > 0 and kind == "ACCT" 
    then   SetFriendNotes(f.acct_characterName,  string.format("%s [rp%x%x%x%x]", note:sub(1, 118), bits))
    elseif kind == "BNET"
    then   BNSetFriendNote(f.bnet_bnetIDAccount, note)
    elseif kind == "ACCT"
    then   SetFriendNotes(f.acct_characterName,  note);
    end;
  end;

  function database:SetCategories(f, set_bits)
    local _, note, bits, kind = self:GetFriendNoteAndCategories(f);
    self:SetFriendNoteAndCategories(f, note, bit.bor(bits, set_bits), kind);
  end;

  function database:AddCategories(f, add_bits)
    local _, note, bits, kind = self:GetFriendNoteAndCategories(f);
    self:SetFriendNoteAndCategories(f, note, bit.bor(bits,add_bits), kind);
  end;

  function database:AddCategory(f, i)
    local _, note, bits, kind = self:GetFriendNoteAndCategories(f);
    self:SetFriendNoteAndCategories(f, note, bit.bor(bits, 2 ^ i), kind);
  end;

  function database:QueryCategory(f, i)        local _, _, bits, _ = self:GetFriendNoteAndCategories(f); return bit.band(bits, 2 ^ i);       end;
  function database:QueryCategories(f, query_bits) local _, _, bits, _ = self:GetFriendNoteAndCategories(f); return bit.band(bits, query_bits or 0); end;

  function database:RemoveCategory(f, i)
    local _, note, bits, kind = self:GetFriendNoteAndCategories(f);
    self:SetFriendNoteAndCategories(f, note, bit.band(bits, bit.bnot(2 ^ i)), kind);
  end;

  function database:RemoveCategories(f, remove_bits)
    local _, note, bits, kind = self:GetFriendNoteAndCategories(f);
    self:SetFriendNoteAndCategories(f, note, bit.band(bits, bit.bnot(remove_bits)), kind);
  end;

  function database:ConvertFriendNote(f)
    local note, kind              = self:PickNote(f);
    local stripped_note, old_cats = note:match("^(.+) %[rp:(%x+)%]"); 
    if   stripped_note
    then local bits = 0;
         for i in old_cats:gmatch("%x") do  bits = bit.bor(bits, 2 ^ tonumber(i, 16)); end;
         self:SetFriendNoteAndCategories(f, stripped_note, bits, kind);
    end;
  end;

  function database:CountCategory(i)
  end;

  function database:ClearCategories(f)
    if f.bnet_noteText     then BNSetFriendNote(f.bnet_bnetIDAccount,     f.bnet_noteText:gsub(" %[rp:%x+%]", "")); end;
    if f.acct_charNoteText then SetFriendNotes(f.acct_characterName,  f.acct_charNoteText:gsub(" %[rp:%x+%]", "")); end;
  end;

  function database:ClearAllCategories()
  end;

  return "data"
end);
